#include <stdio.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>

#include "v4l2_capture.h"

void bgr2rgb(struct capture_info_t *capture_info)
{
int t,c;

  for (t=0; t<capture_info->buffer_len; t=t+3)
  {
    c=capture_info->buffer[t+2];
    capture_info->buffer[t+2]=capture_info->buffer[t];
    capture_info->buffer[t]=c;
  }
}

unsigned char *convert_bayer(struct capture_info_t *capture_info)
{
int x,y,c;
int ptr;

  ptr=0;

  for (y=0; y<capture_info->height; y++)
  {
    c=(y*(capture_info->width));

    for (x=0; x<capture_info->width; x=x+2)
    {
      if ((y%2)==0)
      {
        capture_info->picture[ptr++]=capture_info->buffer[c+capture_info->width+1];
        capture_info->picture[ptr++]=capture_info->buffer[c+capture_info->width];
        capture_info->picture[ptr++]=capture_info->buffer[c];

        capture_info->picture[ptr++]=capture_info->buffer[c+capture_info->width+1];
        capture_info->picture[ptr++]=capture_info->buffer[c+1];
        capture_info->picture[ptr++]=capture_info->buffer[c];
      }
        else
      {
        capture_info->picture[ptr++]=capture_info->buffer[c+1];
        capture_info->picture[ptr++]=capture_info->buffer[c];
        capture_info->picture[ptr++]=capture_info->buffer[c-capture_info->width];

        capture_info->picture[ptr++]=capture_info->buffer[c+1];
        capture_info->picture[ptr++]=capture_info->buffer[c-capture_info->width+1];
        capture_info->picture[ptr++]=capture_info->buffer[c-capture_info->width];
      }

      c=c+2;
    }
  }

  return capture_info->picture;
}

int open_capture(struct capture_info_t *capture_info, char *dev_name)
{
/* char dev_name[]={"/dev/video0"}; */
/* char dev_name[128]; */
char *fourcc;
/* int c; */

  /* sprintf(dev_name,"/dev/video%d",capture_info->device_num); */

  printf("Video Device: %s\n",dev_name);

  /* video_fd=open(dev_name, O_RDONLY, 0); */
  /* video_fd=open(dev_name, O_RDWR | O_NONBLOCK, 0); */

  capture_info->video_fd=open(dev_name, O_RDWR, 0);

  if (capture_info->video_fd==-1)                          
  {
    printf("Problem opening %d\n",capture_info->video_fd);
    return -1;
  }

  if (ioctl(capture_info->video_fd, VIDIOC_QUERYCAP, &capture_info->vid_cap)==-1)
  {
    printf("ioctl error: VIDIOC_QUERYCAP  errno=%d\n",errno); 
    return -1;
  }

#ifdef DEBUG
  printf("      driver: %16s\n",capture_info->vid_cap.driver);
  printf("        card: %32s\n",capture_info->vid_cap.card);
  printf("    bus_info: %32s\n",capture_info->vid_cap.bus_info);
  printf("     version: %d\n",capture_info->vid_cap.version);
  printf("capabilities: 0x%08x\n",capture_info->vid_cap.capabilities);
  printf("\n");
#endif

  if ((capture_info->vid_cap.capabilities&V4L2_CAP_VIDEO_CAPTURE)==0)
  {
    printf("This device doesn't support frame capture\n");
    return -3;
  }

  /* c=capture_info->channel; */
  if (capture_info->channel!=-1)
  { ioctl(capture_info->video_fd, VIDIOC_S_INPUT, &capture_info->channel); }

  /* c=V4L2_STD_NTSC_M; */
  if (capture_info->format!=0)
  { ioctl(capture_info->video_fd, VIDIOC_S_STD, &capture_info->format); }

  memset(&capture_info->vid_fmt,0,sizeof(capture_info->vid_fmt));

  capture_info->vid_fmt.type=V4L2_BUF_TYPE_VIDEO_CAPTURE;
  capture_info->vid_fmt.fmt.pix.width=capture_info->width; 
  capture_info->vid_fmt.fmt.pix.height=capture_info->height;
  // capture_info->vid_fmt.fmt.pix.pixelformat=V4L2_PIX_FMT_RGB24;
  capture_info->vid_fmt.fmt.pix.pixelformat=V4L2_PIX_FMT_BGR24;
  // capture_info->vid_fmt.fmt.pix.pixelformat=V4L2_PIX_FMT_RGB32;
  // capture_info->vid_fmt.fmt.pix.pixelformat=V4L2_PIX_FMT_YUV420;
  capture_info->vid_fmt.fmt.pix.field=V4L2_FIELD_ANY;

  fourcc=(char *)&capture_info->vid_fmt.fmt.pix.pixelformat;
#ifdef DEBUG
  printf(" pixelformat: %08x %c%c%c%c\n",capture_info->vid_fmt.fmt.pix.pixelformat,fourcc[0],fourcc[1],fourcc[2],fourcc[3]);
#endif

  if (ioctl(capture_info->video_fd, VIDIOC_S_FMT, &capture_info->vid_fmt)==-1)
  {
    printf("ioctl error: VIDIOC_S_FMT\n"); 
    return -1;
  }

#ifdef DEBUG
  printf(" pixelformat: %08x %c%c%c%c\n",capture_info->vid_fmt.fmt.pix.pixelformat,fourcc[0],fourcc[1],fourcc[2],fourcc[3]);
  printf(" byteperline: %d\n",capture_info->vid_fmt.fmt.pix.bytesperline);
  printf("   sizeimage: %d\n",capture_info->vid_fmt.fmt.pix.sizeimage);
  printf("  colorspace: %08x\n",capture_info->vid_fmt.fmt.pix.colorspace);
#endif

  if (capture_info->vid_fmt.fmt.pix.pixelformat==V4L2_PIX_FMT_SBGGR8)
  {
    capture_info->picture=(unsigned char *)malloc(capture_info->width*capture_info->height*3);
  }

  capture_info->buffer_len=capture_info->vid_fmt.fmt.pix.sizeimage;
  capture_info->buffer=(unsigned char *)malloc(capture_info->buffer_len);

  return 0;
}

unsigned char *capture_image(struct capture_info_t *capture_info)
{
int c;

  c=0;

  while(c<capture_info->buffer_len)
  {
    c=c+read(capture_info->video_fd, capture_info->buffer+c, (capture_info->buffer_len)-c);
  }

  if (capture_info->vid_fmt.fmt.pix.pixelformat==V4L2_PIX_FMT_BGR24)
  { bgr2rgb(capture_info); }
    else
  if (capture_info->vid_fmt.fmt.pix.pixelformat==V4L2_PIX_FMT_SBGGR8)
  {
    return convert_bayer(capture_info);
  }

  return capture_info->buffer;
}

int close_capture(struct capture_info_t *capture_info)
{
  free(capture_info->buffer);
  if (capture_info->picture!=0) free(capture_info->picture);

  close(capture_info->video_fd);

  return 0;
}



